#include "global.h"


/* Default configuration values */
struct Config cfg = {
    CACHEDIR,				/* cache_directory		*/
    WWWDIR,				/* target_directory		*/
    NULL,				/* reports_site_name		*/
    LOGFILE,				/* log_file			*/
    ETCDIR"/free-sa.conf",		/* configuration_file		*/
    NULL,				/* username_file		*/
    NULL,				/* global_filter		*/
    NULL,				/* users_filter			*/
    NULL,				/* reports_logo			*/
    NULL,				/* configuration_name		*/
    NULL,				/* email_address		*/
    "",					/* locale			*/
    5000L,				/* real_time_timeout		*/
    SAMAXTIMET,				/* reports_rotate		*/
    0L,					/* log_time_zone_shift		*/
    0,					/* log_format			*/
    50,					/* reports_url_limit		*/
    0,					/* top_sites_limit		*/
    0,					/* reports_overwrite		*/
    0,					/* users_graphics_svg		*/
    0,					/* log_skip_errors		*/
    1,					/* users_graphics_svg_mode	*/
    0,					/* users_graphics_svg_pie	*/
    'b',				/* reports_bytes_divisor	*/
    1,					/* username_is_preferred	*/
    0,					/* username_unescape		*/
    0,					/* username_resolve_ip		*/
    1,					/* reports_show_info		*/
    1,					/* reports_indicators		*/
    1,					/* users_fullurl_report		*/
    0,					/* users_fullurl_split		*/
    1,					/* users_report			*/
    1,					/* top_sites_report		*/
    1					/* server_efficiency_report	*/
};

struct owConfig owcfg = {
    NULL,				/* target_directory		*/
    NULL,				/* log_file			*/
    NULL,				/* email_address		*/
    NULL				/* configuration_name		*/
};

struct iConfig icfg = {
    NULL,				/* locDivisor	*/
    NULL,				/* ReportDir	*/
    NULL,				/* CharSet	*/
    NULL,				/* locFile	*/
    ".total",				/* TTLname	*/
    ".average",				/* AVGname	*/
    "description",			/* DFname	*/
    "index",				/* IRname	*/
    "index",				/* LRname	*/
    "topuser",				/* TURname	*/
    "topsites",				/* TSRname	*/
    "email",				/* ERname	*/
    "server_efficiency",		/* SERname	*/
    0,					/* SYear	*/
    0,					/* EYear	*/
    0,					/* LFcount	*/
    0,					/* LFcsscount	*/
    0L,					/* SDate	*/
    SAMAXTIMET,				/* EDate	must be high enough */
    0,					/* LocDateLen	*/
    0,					/* dldCount	*/
    0,					/* utCount	*/
    0,					/* ueCount	*/
    1.0,				/* reports_bytes_divisor	*/
    NULL,				/* Downloads	*/
    NULL,				/* Usertab	*/
    "\000",				/* Period	*/
    SAB0,				/* SAB		*/
    0,					/* GF		*/
    0,					/* UF		*/
    0,					/* Mode		*/
    31					/* PScale	*/
};

struct Localisation defloc = {
    "No",				/* No			*/
    "User",				/* User			*/
    "Requests",				/* Connections		*/
    "Seconds",				/* Seconds		*/
    "Bytes",				/* Bytes		*/
    "URL",				/* URL			*/
    "Period",				/* Period		*/
    "Sort field (order)",		/* SortFieldNOrder	*/
    "Users",				/* Users		*/
    "Served",				/* Served		*/
    "Actual",				/* Actual		*/
    "Cashed",				/* Cashed		*/
    "Server denied (ACL)",		/* Denied		*/
    "Traffic type",			/* TrafType		*/
    "Other local",			/* OthLocal		*/
    "Top user",				/* TopUser		*/
    "Top sites",			/* TopSites		*/
    "Server efficiency",		/* SEfficiency		*/
    " in real time",			/* InRealTime		*/
    "Date and Time",			/* DnT			*/
    "Date\\Time",			/* DsT			*/
    "Day of month",			/* DoM			*/
    "asc.",				/* Asc			*/
    "desc.",				/* Desc			*/
    "TOTAL",				/* TOTAL		*/
    "AVERAGE",				/* AVERAGE		*/
    "MAXIMUM",				/* MAXIMUM		*/
    "Code",				/* Code			*/
    "Generated by",			/* GeneratedBy		*/
    "Configuation",			/* Configuration	*/
    "Generation date",			/* GenerationD		*/
    "Period selection page",		/* PSP			*/
    "P",				/* P			*/
    "T",				/* T			*/
    "G",				/* G			*/
    "M",				/* M			*/
    "k",				/* k			*/
    "Sender",				/* Sender		*/
    "Messages",				/* Messages		*/
    "Recipient",			/* Recipient		*/
    "Senders",				/* Senders		*/
    "Recipients"			/* Recipients		*/
};

struct RConfig rcfg = {
    NULL,                              /* UFRsort              */
    NULL,                              /* URsort               */
    NULL,                              /* UDRsort              */
    NULL                               /* LFRsort              */
};


struct Localisation *loc = &defloc;
struct FConfig *gfcfg = NULL;
struct FConfig *ufcfg = NULL;
struct FConfig **lfcfgs = NULL;


short int CS2SI(char *ivalue, short int srange, short int erange);
short int CS2SI(char *ivalue, short int srange, short int erange) {
    short int retval;
    
    if (sscanf(ivalue, "%hd", &retval) != 1) SAexit("Error converting \"%s\" to short int", ivalue);
    if ((srange != -1 && retval < srange) || (erange != -1 && retval > erange)) SAexit("Value %hd is out of range from %hd to %hd", retval, srange, erange);
    
    return retval;
}


double CS2D(char *ivalue);
double CS2D(char *ivalue) {
    double retval;
    
    if (sscanf(ivalue, "%lf", &retval) != 1) SAexit("Error converting \"%s\" to double", ivalue);
    
    return retval;
}


time_t CS2TT(char *ivalue);
time_t CS2TT(char *ivalue) {
    time_t retval;

    if (sscanf(ivalue, "%10ld", &retval) != 1)
	if (!(retval = GetDW(ivalue))) SAexit("Error converting \"%s\" to time_t", ivalue);
	
    return retval;
}


char CS2B(char *ivalue);
char CS2B(char *ivalue) {
    if (strchr("tTyYeE1", ivalue[0]) != NULL) return 1;
    else if (strchr("fFnNdD0", ivalue[0]) == NULL) SAexit("Error converting \"%s\" to bool", ivalue);
    
    return 0;
}


char CS2C(char *ivalue);
char CS2C(char *ivalue) {
    if (ivalue[1] != '\0') SAexit("Error converting \"%s\" to char", ivalue);

    return ivalue[0];
}


void LoadUsertab(void);
void LoadUsertab(void) {
    char *tBuf, *t;

    if ((tBuf = (char *)SAfreadMC(NULL, cfg.username_file)) == NULL) SAexit("Can't stat file");
    
    while((t = strchr(tBuf, ' ')) != NULL) {
	icfg.Usertab = (struct SAusertab *)SArealloc(icfg.Usertab, sizeof(struct SAusertab) * ((size_t)icfg.utCount + 1L));
	icfg.Usertab[icfg.utCount].iname = tBuf;
	t[0] = '\0';
	tBuf = t + 1L;
	
	if ((t = strchr(tBuf, '\n')) == NULL) SAexit("Incorrect file format");
	t[0] = '\0';
	icfg.Usertab[icfg.utCount].ename = tBuf;
	tBuf = t + 1L;
	
#if DEBUG >= 1
	SAdebug("%hd\t%s->%s", icfg.utCount + 1, icfg.Usertab[icfg.utCount].iname, icfg.Usertab[icfg.utCount].ename);
#endif

	icfg.utCount++;
    }

    if (icfg.utCount) qsort(icfg.Usertab, (size_t)icfg.utCount, sizeof(struct SAusertab), (int (*)(const void *, const void *))compSAusertab);
    else SAwarning("Empty file");
}


void LoadLocalisation(void);
void LoadLocalisation(void) {
    char *tBuf, *t;
    static char *newloc[LOCSIZE];
    short int locCount = 0;

    if ((tBuf = (char *)SAfreadMC(NULL, icfg.locFile)) == NULL) SAexit("Can't stat file");
    
    while((t = strchr(tBuf, '\n')) != NULL) {
	newloc[locCount] = tBuf;
	t[0] = '\0';
	tBuf = t + 1L;

#if DEBUG >= 1
	SAdebug("%hd\t%s", locCount + 1, newloc[locCount]);
#endif

	locCount++;
    }

    if (locCount != LOCSIZE) SAexit("Incorrect file format");
    
    loc = (struct Localisation *)((char **)(&newloc));
}


void SAfilterrxcomp(void ***rxchar, short int iCount);
void SAfilterrxcomp(void ***rxchar, short int iCount) {
    short int i;
    regex_t *nrxchar = (regex_t *)SAmalloc(sizeof(regex_t) * (size_t)iCount);

    /* compile regexps */
    for (i = 0; i < iCount; i++) if (regcomp(&nrxchar[i], rxchar[0][i], REG_EXTENDED)) SAexit("Incorrect regular expression '%s'", rxchar[0][i]);
    
    /* reassign pointer to compiled regexps */
    free(*rxchar);
    *rxchar = (void **)nrxchar;
}


struct FConfig *LoadFilter(const char *fFile);
struct FConfig *LoadFilter(const char *fFile) {
    char *tBuf, *t;
    struct FConfig *fcfg, tfcfg = {		/* Initial values	*/
	0,					/* inameCount		*/
	0,					/* urlCount		*/
	0,					/* ipCount		*/
        0,					/* codeCount		*/
        50,					/* limit		*/
	NULL,					/* inames		*/
        NULL,					/* urls			*/
        NULL,					/* ips			*/
        NULL,					/* codes		*/
        NULL,					/* methods		*/
        NULL,					/* name			*/
        '\0',					/* stats		*/
        0.0,					/* bytes		*/
        0,					/* inameMode		*/
	0,					/* urlMode		*/
	0,					/* ipMode		*/
        0,					/* codeMode		*/
        0,					/* methodMode		*/
        0,					/* statMode		*/
        0,					/* bytesMode		*/
        1					/* showbytes		*/
    };
    
    
    if ((tBuf = (char *)SAfreadMC(NULL, fFile)) == NULL) SAexit("Can't stat file '%s'", fFile);

    /* Init filter */
    fcfg = (struct FConfig *)malloc(sizeof(struct FConfig));
    /*memcpy(fcfg, &tfcfg, sizeof(struct FConfig));*/
    fcfg[0] = tfcfg;
    
    for(; (t = strchr(tBuf, '\n')) != NULL; tBuf = t + 1L) {
	/* Skip comments and empty lines */
	if (tBuf[0] == '\n' || tBuf[0] == '#') continue;
	/* Check for correct format */
	if (tBuf[1] != ' ') goto fferror;
	/* Terminate line */
	t[0] = '\0';
	switch(tBuf[0]) {
	    case 'a':
		    fcfg->ips = (void **)SArealloc(fcfg->ips, sizeof(char *) * (fcfg->ipCount + 1L));
		    fcfg->ips[fcfg->ipCount] = &tBuf[2];
		    fcfg->ipCount++;
		    break;
	    case 'A':
		    if (tBuf[2] >= '0' && tBuf[2] <= '5') fcfg->ipMode = tBuf[2] - '0';
		    else goto fferror;
		    break;
	    case 'b':
		    fcfg->bytes = CS2D(&tBuf[2]);
		    break;
	    case 'B':
		    if (tBuf[2] >= '0' && tBuf[2] <= '1') fcfg->bytesMode = tBuf[2] - '0';
		    else goto fferror;
		    break;
	    case 'c':
		    fcfg->codes = (short int *)SArealloc(fcfg->codes, sizeof(short int) * (fcfg->codeCount + 1L));
		    fcfg->codes[fcfg->codeCount] = CS2SI(&tBuf[2], 0, CODESSIZE);
		    fcfg->codeCount++;
		    break;
	    case 'C':
		    if (tBuf[2] >= '0' && tBuf[2] <= '1') fcfg->codeMode = tBuf[2] - '0';
		    else goto fferror;
		    break;
	    case 'i':
		    fcfg->inames = (void **)SArealloc(fcfg->inames, sizeof(char *) * (fcfg->inameCount + 1L));
		    fcfg->inames[fcfg->inameCount] = &tBuf[2];
		    fcfg->inameCount++;
		    break;
	    case 'I':
		    if (tBuf[2] >= '0' && tBuf[2] <= '5') fcfg->inameMode = tBuf[2] - '0';
		    else goto fferror;
		    break;
	    case 'l':
		    fcfg->limit = CS2SI(&tBuf[2], 0, -1);
		    break;
	    case 'm':
		    if ((!isupper(tBuf[2])) || tBuf[3] != '\0') goto fferror;
		    if (fcfg->methods == NULL) fcfg->methods = SAstrdup(&tBuf[2]);
		    else {
			fcfg->methods = (char *)SArealloc(fcfg->methods, sizeof(char) * (strlen(fcfg->methods) + 2L));
			strncat(fcfg->methods, &tBuf[2], 1L);
		    }
		    break;
	    case 'M':
		    if (tBuf[2] >= '0' && tBuf[2] <= '1') fcfg->methodMode = tBuf[2] - '0';
		    else goto fferror;
		    break;
	    case 'n':
		    fcfg->name = &tBuf[2];
		    break;
	    case 's':
		    if (tBuf[2] >= '0' && tBuf[2] <= '3') SAsetbit(fcfg->stats, (tBuf[2] - '0'));
		    else goto fferror;
		    break;
	    case 'S':
		    if (tBuf[2] >= '0' && tBuf[2] <= '1') fcfg->statMode = tBuf[2] - '0';
		    else goto fferror;
		    break;
	    case 'u':
		    fcfg->urls = (void **)SArealloc(fcfg->urls, sizeof(char *) * (fcfg->urlCount + 1L));
		    fcfg->urls[fcfg->urlCount] = &tBuf[2];
		    fcfg->urlCount++;
		    break;
	    case 'U':
		    if (tBuf[2] >= '0' && tBuf[2] <= '5') fcfg->urlMode = tBuf[2] - '0';
		    else goto fferror;
		    break;
	    case 'w':
		    if (tBuf[2] >= '0' && tBuf[2] <= '1') fcfg->showbytes = tBuf[2] - '0';
		    else goto fferror;
		    break;
	    default:
fferror:
		    SAexit("Incorrect file format '%s'", fFile);
		    break;
	}
#if DEBUG >= 1
	SAdebug("%c\t%s", tBuf[0], &tBuf[2]);
#endif
    }

    if (fcfg->inameCount || fcfg->urlCount || fcfg->ipCount || fcfg->codeCount || fcfg->methods != NULL || fcfg->stats != '\0' || fcfg->bytes != 0.0) {
	if (fcfg->codeCount) qsort(fcfg->codes, (size_t)fcfg->codeCount, sizeof(short int), (int (*)(const void *, const void *))compSI);
	if (fcfg->inameMode == 2 || fcfg->inameMode == 3) qsort(fcfg->inames, (size_t)fcfg->inameCount, sizeof(char *), (int (*)(const void *, const void *))SApstrcmp);
	else if (fcfg->inameMode == 4 || fcfg->inameMode == 5) SAfilterrxcomp(&fcfg->inames, fcfg->inameCount);
	if (fcfg->urlMode == 2 || fcfg->urlMode == 3) qsort(fcfg->urls, (size_t)fcfg->urlCount, sizeof(char *), (int (*)(const void *, const void *))SApstrcmp);
	else if (fcfg->urlMode == 4 || fcfg->urlMode == 5) SAfilterrxcomp(&fcfg->urls, fcfg->urlCount);
	if (fcfg->ipMode == 2 || fcfg->ipMode == 3) qsort(fcfg->ips, (size_t)fcfg->ipCount, sizeof(char *), (int (*)(const void *, const void *))SApstrcmp);
	else if (fcfg->ipMode == 4 || fcfg->ipMode == 5) SAfilterrxcomp(&fcfg->ips, fcfg->ipCount);
	
	if (fcfg->name == NULL) fcfg->name = SAstrdup(fFile);
	
	return fcfg;
    }
    
    free(fcfg);
    SAwarning("Empty file '%s'", fFile);
    return NULL;
}


void SAusage(char *iName, char mode, char *itext);
void SAusage(char *iName, char mode, char *itext) {
    const char *const utext = "\
Usage: %s [-s|-r|-F|-h|-v|-V] [-d STARTDATE-ENDDATE] [-f FILE] [-l FILE] [-o DIRECTORY] [-n NAME]\n\
  -d STARTDATE-ENDDATE	Dates range for report generation in current locale,\n\
			you may specify only STARTDATE followed by '-' sign,\n\
			or only ENDDATE preceded by '-' sign.\n\
  -e ADDRESS		Send topuser report in text format to e-mail ADDRESS,\n\
			if ADDRESS is '-' then report is writed to stdout.\n\
  -f FILE		Read configuration from FILE.\n\
  -F			Fix root index.html file according to\n\
			reports directories presense.\n\
  -h			Show help message\n\
  -l FILE		Use FILE as log file.\n\
  -n NAME		Use NAME as configuration name.\n\
  -o DIRECTORY		Use DIRECTORY as target for placing reports.\n\
  -r			Remove records from log file.\n\
			You must also specify date(s) at '-d' option.\n\
  -s			Show log file information:\n\
			name, size, start and end dates.\n\
  -v			Show version\n\
  -V			Show advanced information about current "PROGLNAME" build\n";

    char etext[URLSIZE];
    int elen = 0;


    if (mode == 0) {
	SAshow(utext, iName);
	exit(EXIT_SUCCESS);
    } else {
	switch(mode) {
	    case 1:
		    elen = SAsnprintf(etext, sizeof(etext), "Incorrect option '%s'\n", itext);
		    break;
	    case 2:
		    elen = SAsnprintf(etext, sizeof(etext), "Incorrect dates range specified '%s'\n", itext);
		    break;
	    case 3:
		    elen = SAsnprintf(etext, sizeof(etext), "Option '%s' requires argument\n", itext);
		    break;
	    case 4:
		    elen = SAsnprintf(etext, sizeof(etext), "Too much action options specified\n");
		    break;
	}
	
	SAsnprintf(&etext[elen], URLSIZE - (long)elen, utext, iName);
	SAexit(etext);
    }
}


void SAversion(char fullinfo);
void SAversion(char fullinfo) {
    SAshow(
PROGNAME" "PROGVERSION" ("PROGURL")\n"PROGCOPYRIGHT"\n\
License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n\
This program comes with ABSOLUTELY NO WARRANTY.\n\
This is free software, and you are welcome to redistribute it under certain conditions.\n\
For details please read the file named COPYING.\n");

    if (fullinfo) SAshow("\n\
PREFIX			"PREFIX"\n\
ETCDIR			"ETCDIR"\n\
WWWDIR			"WWWDIR"\n\
CACHEDIR		"CACHEDIR"\n\
LOCDIR			"LOCDIR"\n\
DOCDIR			"DOCDIR"\n\
LOGFILE			"LOGFILE"\n\
CFLAGS			"CFLAGS"\n\
LDFLAGS			"LDFLAGS"\n\
LIBS			"LIBS"\n\
Thousands separator	%s\n\
Large file support	%s\n\
Debug level		%d\n\
C compiler		"CC"\n\
Install command		"INSTALL"\n\
Sort command		"CMDSORT"\n", 
			(SATSEP[0] == '\'') ? "yes" : "no", (LARGEFILE == 0) ? "no" : "yes", DEBUG);

    exit(EXIT_SUCCESS);
}


void InitConfigWithCL(int argc, char **argv) {
    short int i = 1;
    time_t ttime;

    
    /* Parse command line options */
    if (argc > 1)
	for (; i < argc; i++) {
	    if (argv[i][0] != '-' || argv[i][2] != '\0') SAusage(argv[0], 1, argv[i]);
	    switch(argv[i][1]) {
		case 'd':
			if (i == argc - 1) SAusage(argv[0], 3, argv[i]);
			if (islower(argv[i + 1][0])) {	/* assume date as one of predefined words */
			    if (time(&icfg.EDate) == -1L) SAexit("Can't get current date. %s", strerror(errno));
			    if (!(ttime = GetDW(argv[i + 1]))) SAusage(argv[0], 2, argv[i + 1]);
			    icfg.SDate = icfg.EDate - ttime;
			} else {			/* assume date in locale format */
			    if (argv[i + 1][0] == '-') icfg.EDate = GetDZT(&argv[i + 1][1]) + SADAY - 1L;
			    else {
			        if (argv[i + 1][icfg.LocDateLen] != '-') SAusage(argv[0], 2, argv[i + 1]);
			        argv[i + 1][icfg.LocDateLen] = '\0';
			        icfg.SDate = GetDZT(&argv[i + 1][0]);
			        if (argv[i + 1][icfg.LocDateLen + 1] != '\0') icfg.EDate = GetDZT(&argv[i + 1][icfg.LocDateLen + 1]) + SADAY - 1L;
			    }
			}
			i++; /* 2 args readed for this option */
			break;
		case 'e':
			if (i == argc - 1) SAusage(argv[0], 3, argv[i]);
			owcfg.email_address = argv[i + 1];
			i++; /* 2 args readed for this option */
			break;
		case 'f':
			if (i == argc - 1) SAusage(argv[0], 3, argv[i]);
			cfg.configuration_file = argv[i + 1];
			i++; /* 2 args readed for this option */
			break;
		case 'h':
			SAusage(argv[0], 0, NULL);
			break;
		case 'l':
			if (i == argc - 1) SAusage(argv[0], 3, argv[i]);
			owcfg.log_file = argv[i + 1];
			i++; /* 2 args readed for this option */
			break;
		case 'n':
			if (i == argc - 1) SAusage(argv[0], 3, argv[i]);
			owcfg.configuration_name = argv[i + 1];
			i++; /* 2 args readed for this option */
			break;
		case 'o':
			if (i == argc - 1) SAusage(argv[0], 3, argv[i]);
			owcfg.target_directory = argv[i + 1];
			i++; /* 2 args readed for this option */
			break;
		case 'r':
			if (icfg.Mode != 0) SAusage(argv[0], 4, NULL);
			else icfg.Mode = 3;
			break;
		case 's':
			if (icfg.Mode != 0) SAusage(argv[0], 4, NULL);
			else icfg.Mode = 2;
			break;
		case 'v':
			SAversion(0);
			break;
		case 'V':
			SAversion(1);
			break;
		case 'F':
			if (icfg.Mode != 0) SAusage(argv[0], 4, NULL);
			else icfg.Mode = 4;
			break;
		default:
			SAusage(argv[0], 1, argv[i]);
			break;
	    }
	}

    InitConfig();
}


void InitConfig(void) {
    FILE *CONF;
    char tBuf[128];
    char *value;
    double dpow = 0.0;
    struct FConfig *fcfg;
    
    
    /* Open configuration file */
    CONF = SAfopenI("r", cfg.configuration_file);
    
    /* Read and parse configuration file */
    while (fgets(tBuf, 128, CONF) != NULL) {
	/* Skip comments and empty lines */
	if (tBuf[0] == '\n' || tBuf[0] == '#') continue;
	
	/* Split for parameter and value at =, remove " from value */
	if ((value = strchr(tBuf, '=')) == NULL) SAexit("Error parsing confuguration file, line was: %s", tBuf);
	value[0] = '\0';
	value += 2L;
	value[strlen(value) - 2L] = '\0';
#if DEBUG >= 1	
	SAdebug("%s=\"%s\"", tBuf, value);
#endif
	/* Empty value */
	if (value[0] == '\0') continue;
	
	switch(tBuf[0]) {
	    case 'c':
			 if (!strcmp(tBuf, "cache_directory"))		cfg.cache_directory = SAstrdup(value);
		    else if (!strcmp(tBuf, "configuration_name"))	cfg.configuration_name = SAstrdup(value);
		    else goto opterror;
		    break;
	    case 'e':
		    if (!strcmp(tBuf, "email_address"))			cfg.email_address = SAstrdup(value);
		    else goto opterror;
		    break;
	    case 'g':
		    if (!strcmp(tBuf, "global_filter"))			cfg.global_filter = SAstrdup(value);
		    else goto opterror;
		    break;
	    case 'l':
			 if (!strcmp(tBuf, "locale"))			cfg.locale = SAstrdup(value);
		    else if (!strcmp(tBuf, "local_filter")) {
			if ((fcfg = LoadFilter(value)) != NULL) {
			    if (icfg.LFcount == MAXLFCOUNT) {
				SAwarning("Filter '%s' will be skipped. Only %d local filters are supported now, please contact developers if you need more", value, MAXLFCOUNT);
				break;
			    }
			    icfg.LFcount++;
			    lfcfgs = (struct FConfig **)SArealloc(lfcfgs, sizeof(struct FConfig *) * (size_t)icfg.LFcount);
			    lfcfgs[icfg.LFcount - 1L] = fcfg;
			}
		    }
		    else if (!strcmp(tBuf, "log_file"))			cfg.log_file = SAstrdup(value);
		    else if (!strcmp(tBuf, "log_format"))		cfg.log_format = CS2SI(value, 0, (short int)(sizeof(LogFormats)/sizeof(struct LogFormat)) - 1);
		    else if (!strcmp(tBuf, "log_skip_errors"))		cfg.log_skip_errors = CS2B(value);
		    else if (!strcmp(tBuf, "log_time_zone_shift"))	cfg.log_time_zone_shift = CS2TT(value);
		    else goto opterror;
		    break;
	    case 'r':
			 if (!strcmp(tBuf, "real_time_timeout"))	cfg.real_time_timeout = CS2TT(value);
		    else if (!strcmp(tBuf, "reports_bytes_divisor"))	cfg.reports_bytes_divisor = CS2C(value);
		    else if (!strcmp(tBuf, "reports_indicators"))	cfg.reports_indicators = CS2B(value);
		    else if (!strcmp(tBuf, "reports_logo"))		cfg.reports_logo = SAstrdup(value);
		    else if (!strcmp(tBuf, "reports_overwrite"))	cfg.reports_overwrite = CS2SI(value, 0, 2);
		    else if (!strcmp(tBuf, "reports_rotate"))		cfg.reports_rotate = CS2TT(value);
		    else if (!strcmp(tBuf, "reports_show_info"))	cfg.reports_show_info = CS2B(value);
		    else if (!strcmp(tBuf, "reports_site_name"))	cfg.reports_site_name = SAstrdup(value);
		    else if (!strcmp(tBuf, "reports_url_limit"))	cfg.reports_url_limit = CS2SI(value, 0, -1);
		    else goto opterror;
		    break;
	    case 's':
		    if (!strcmp(tBuf, "server_efficiency_report"))	cfg.server_efficiency_report = CS2B(value);
		    else goto opterror;
		    break;
	    case 't':
		         if (!strcmp(tBuf, "target_directory"))		cfg.target_directory = SAstrdup(value);
		    else if (!strcmp(tBuf, "top_sites_limit"))		cfg.top_sites_limit = CS2SI(value, 0, -1);
		    else if (!strcmp(tBuf, "top_sites_report"))		cfg.top_sites_report = CS2B(value);
		    else goto opterror;
		    break;
	    case 'u':
			 if (!strcmp(tBuf, "username_file"))		cfg.username_file = SAstrdup(value);
		    else if (!strcmp(tBuf, "username_is_preferred"))	cfg.username_is_preferred = CS2B(value);
		    else if (!strcmp(tBuf, "username_resolve_ip"))	cfg.username_resolve_ip = CS2B(value);
		    else if (!strcmp(tBuf, "username_unescape"))	cfg.username_unescape = CS2B(value);
		    else if (!strcmp(tBuf, "users_filter"))		cfg.users_filter = SAstrdup(value);
		    else if (!strcmp(tBuf, "users_fullurl_report"))	cfg.users_fullurl_report = CS2B(value);
		    else if (!strcmp(tBuf, "users_fullurl_split"))	cfg.users_fullurl_split = CS2B(value);
		    else if (!strcmp(tBuf, "users_graphics_svg"))	cfg.users_graphics_svg = (short int)CS2B(value);
		    else if (!strcmp(tBuf, "users_graphics_svg_mode"))	cfg.users_graphics_svg_mode = CS2B(value);
		    else if (!strcmp(tBuf, "users_graphics_svg_pie"))	cfg.users_graphics_svg_pie = CS2B(value);
		    else if (!strcmp(tBuf, "users_report"))		cfg.users_report = CS2B(value);
		    else goto opterror;
		    break;
	    default:
opterror:
		    SAexit("Unknown parameter '%s' with value '%s'", tBuf, value);
		    break;
	}
    }

    /* Close configuration file */
    SAfclose(CONF);

    /* Command line options arguments is preferred over config file */
    if (owcfg.target_directory != NULL) cfg.target_directory = owcfg.target_directory;
    if (owcfg.log_file != NULL) cfg.log_file = owcfg.log_file;
    if (owcfg.email_address  != NULL) cfg.email_address  = owcfg.email_address;
    if (owcfg.configuration_name != NULL) cfg.configuration_name = owcfg.configuration_name;
    /* If no configuration name specified, then use config file name */
    else if (cfg.configuration_name == NULL) cfg.configuration_name = cfg.configuration_file;

    /* If alternative mode not selected at command line, then choose log processing */
    if (icfg.Mode == 0)	icfg.Mode = 1;
    /* Otherwise disable overwrite mode, because we will not generate reports */
    else cfg.reports_overwrite = 0;

    /* If alternative locale specified, reinitialise it */
    if (cfg.locale[0] != '\0') InitLocale();

    /* Load files, which configuration depend */
    if (cfg.username_file != NULL) LoadUsertab();
    if (icfg.locFile != NULL) LoadLocalisation();
    if (cfg.global_filter != NULL) icfg.GF = (((gfcfg = LoadFilter(cfg.global_filter)) == NULL) ? 0 : 1);
    if (cfg.users_filter != NULL) icfg.UF = (((ufcfg = LoadFilter(cfg.users_filter)) == NULL) ? 0 : 1);
    /* Compute number of local filters which require indicators
       and disable indicators if there are no local filters at all */
    if (!(icfg.LFcsscount = SAmin(icfg.LFcount, MAXLFCSSCOUNT))) cfg.reports_indicators = 0;
    
    if (cfg.users_fullurl_report && !cfg.users_report) SAexit("In order to enable users report with full URLs you should also enable users report which is disabled now.");
    
    /* Compute result of bytes division name AFTER InitLocale */
    switch (cfg.reports_bytes_divisor) {
        case 'b':
        case 'B': icfg.SAB = SAB0; dpow =  0.0; break;
        case 'k': icfg.SAB = SAB1; dpow = 10.0; icfg.locDivisor = loc->k; break;
        case 'K': icfg.SAB = SAB2; dpow = 10.0; icfg.locDivisor = loc->k; break;
        case 'm': icfg.SAB = SAB1; dpow = 20.0; icfg.locDivisor = loc->M; break;
        case 'M': icfg.SAB = SAB2; dpow = 20.0; icfg.locDivisor = loc->M; break;
        case 'g': icfg.SAB = SAB1; dpow = 30.0; icfg.locDivisor = loc->G; break;
        case 'G': icfg.SAB = SAB2; dpow = 30.0; icfg.locDivisor = loc->G; break;
        case 't': icfg.SAB = SAB1; dpow = 40.0; icfg.locDivisor = loc->T; break;
        case 'T': icfg.SAB = SAB2; dpow = 40.0; icfg.locDivisor = loc->T; break;
        case 'p': icfg.SAB = SAB1; dpow = 50.0; icfg.locDivisor = loc->P; break;
        case 'P': icfg.SAB = SAB2; dpow = 50.0; icfg.locDivisor = loc->P; break;
        case 'v':
        case 'V': icfg.SAB = SAB3; dpow =  0.0; break;
        default: SAexit("Incorrect divisor specified: '%c'", cfg.reports_bytes_divisor); break;
    }
    icfg.reports_bytes_divisor = pow(2.0,  dpow);
    
    SAB3variants[0].munit = loc->P;
    SAB3variants[1].munit = loc->T;
    SAB3variants[2].munit = loc->G;
    SAB3variants[3].munit = loc->M;
    SAB3variants[4].munit = loc->k;
}


void InitLocale(void) {
    static char locstring[FILENAMESIZE];
    const char *t = NULL;
    char *curlocale;
    short int i;

    
    /* Init locale */
    curlocale = SAsetlocale(LC_ALL, cfg.locale);
    icfg.CharSet = nl_langinfo(CODESET);
    
    /* Compute crossplatform charset and fill with it localisation file name only if we not in C locale */
    if ((curlocale[0] != 'C' && curlocale[1] != '\0') && (curlocale[0] != 'L' && curlocale[1] != 'C')) {
	i = SAsnprintf(locstring, sizeof(locstring), LOCDIR"/%c%c.", curlocale[0], curlocale[1]);
	for (t = icfg.CharSet; t[0] != '\0'; t++) {
	    if (t[0] == '-') continue;
	    locstring[i++] = (char)tolower((int)t[0]);
	}
	locstring[i] = '\0';
	icfg.locFile = SAstrdup(locstring);
#if DEBUG >= 1
	SAdebug("Selected localisation file: %s", icfg.locFile);
#endif
    }
#if DEBUG >= 1
    else SAdebug("Selected localisation file: internal English");
#endif
    
    /* Export variables for spawned programs */
    SAsnprintf(locstring, sizeof(locstring), "LC_ALL=%s", cfg.locale);
    if (putenv(locstring)) SAexit("Can't set LC_ALL to %s", locstring);
    if (putenv("LC_COLLATE=POSIX")) SAexit("Can't set LC_COLLATE to POSIX");

    /* Compute max date length */    
    icfg.LocDateLen = strlen(GetDT(SALONGDATE, "%x"));
}
