#include "global.h"


double x = 960.0, y;


int compCDentry(const struct ChartData *el1, const struct ChartData *el2);
int compCDentry(const struct ChartData *el1, const struct ChartData *el2) {
    return SAnumcmp(el1->value, el2->value);
}


FILE *SAfopenS(void);
FILE *SAfopenS(void) {
    FILE *retval;
    char tfname[FILENAMESIZE];
    
    SAsnprintf(tfname, sizeof(tfname), "%s/%d.svg", icfg.ReportDir, cfg.users_graphics_svg++);
    retval = SAfopenI("w", tfname);
    
    SAfprintf(retval, "\
<?xml version=\"1.0\" encoding=\"KOI8-R\"?>\n\
<?xml-stylesheet href=\"../gsa.css\" type=\"text/css\"?>\n\
<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\" \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">\n\
\t<svg viewBox=\"0 0 %.0f %.0f\" xmlns=\"http://www.w3.org/2000/svg\" version=\"1.1\" xmlns:xlink=\"http://www.w3.org/1999/xlink\">\n",
              x, y);
    
    return retval;
}


void SAfcloseS(FILE *ifile);
void SAfcloseS(FILE *ifile) {
    SAfwrite1("</svg>\n", 7L, ifile);
    SAfclose(ifile);
}


int SSVGcountnmax(struct ChartData *icdata, double *maxvalue);
int SSVGcountnmax(struct ChartData *icdata, double *maxvalue) {
    int retval = 0;

    *maxvalue = 0.0;
    for (; icdata[retval].label != NULL; retval++) *maxvalue = SAmax(*maxvalue, icdata[retval].value);

    return retval;
}


int SSVGcountnsum(struct ChartData *icdata, double *sumvalue);
int SSVGcountnsum(struct ChartData *icdata, double *sumvalue) {
    int retval = 0;

    *sumvalue = 0.0;
    for (; icdata[retval].label != NULL; retval++) *sumvalue += icdata[retval].value;

    return retval;
}


void SSVGdrawpie(struct ChartData *icdata, const char legend);
void SSVGdrawpie(struct ChartData *icdata, const char legend) {
    FILE *ostream;
    int count, i, j;
    double c_y0, c_x0, c_R, c_fontsize, c_rfontindent, c_fontindent, c_pi, c_linewidth, c_barwidth, c_ibspace,
    percent, sxpos, sypos, expos, eypos, xpos, ypos, rad, tmp, sumvalue, sum = 0.0;
    const char *label, *llabel = "*";
    short int ind;


    /* compute max value and count */
    count = SSVGcountnsum(icdata, &sumvalue);
    c_pi = 2.0 * SSVGPI / sumvalue;					/* this is useful constant	*/
    y = x / YXRATIO;
    c_fontsize = y / FSDIV * PC_FSCALE;					/* size of font			*/
    c_R = (y - c_fontsize * (5.0 * TINDENT + 3.0)) / 2.0;		/* radius of circle		*/
    if (legend) c_x0 = x / 4.0;						/* x of circle center		*/
    else c_x0 = x / 2.0;
    c_y0 = c_R + c_fontsize * (2.0 * TINDENT + 1.0);			/* y of circle center		*/
    c_linewidth = y / LWDIV;						/* width of line		*/
    c_fontindent = c_fontsize * TINDENT;				/* font indent			*/
    c_rfontindent = c_R + c_fontindent;					/* font indent from center	*/
    
    qsort(icdata, (size_t)count, sizeof(struct ChartData), (int (*)(const void *, const void *))compCDentry);

    ostream = SAfopenS();
    
    /* initial values */
    sxpos = c_x0 + c_R;
    sypos = c_y0;
    rad = 0.0;
    
    for (i = 0; i < count && icdata[i].value; i++) {
	if ((tmp = c_pi * icdata[i].value) > c_pi * sumvalue / 2.0) ind = 1;
	else ind = 0;
	
	/* sectors calculations */
	rad += tmp;
	/* not show pies smaller than PC_MAXPIE * 100% */
	if (icdata[i].value / sumvalue < PC_MAXPIE && i != count - 1) {
	    rad = 0.0;							/* no space for other sectors	*/
	    percent = 100.0 - sum;
	    label = llabel;
	} else {
	    percent = icdata[i].value * 100.0 / sumvalue;
	    sum += percent;
	    label = icdata[i].label;
	}
	expos = c_x0 + cos(rad) * c_R;
	eypos = c_y0 + sin(rad) * c_R;
	/* labels calculations */
	tmp = rad - tmp / 2.0;
	xpos = c_x0 + cos(tmp) * c_rfontindent;
	ypos = c_y0 + sin(tmp) * c_rfontindent;
	if (ypos > c_y0) ypos += c_fontsize;

	/* workaround */
	if (percent == 100.0) eypos -= 0.0001;

	/* draw sectors and labels */
	SAfprintf(ostream, "\
\t\t<path class=\"c%d\" d=\"M%lf,%lf L%lf,%lf A%lf,%lf 0 %hd,1 %lf,%lf z\" stroke-width=\"%lf\"/>\n\
\t\t<text class=\"%ctext\" x=\"%lf\" y=\"%lf\" font-size=\"%lf\">%s</text>\n\
\t\t<text x=\"%lf\" y=\"%lf\" font-size=\"%lf\">%.1lf%%</text>\n",
		(i % 7) + 1, c_x0, c_y0, sxpos, sypos, c_R, c_R, ind, expos, eypos, c_linewidth,
		(xpos > c_x0) ? 's' : 'e', xpos, ypos, c_fontsize, label,
		(xpos - c_x0) * PC_TINDENTI + c_x0, (ypos - c_y0) * PC_TINDENTI + c_y0 + c_fontsize / 4.0, c_fontsize / 2.0, percent);
		
	if (rad == 0.0) break;
	
	sxpos = expos;
	sypos = eypos;
    }
    
    /* draw summary label */
    SAfprintf(ostream, "\
\t\t<text x=\"%lf\" y=\"%lf\" font-size=\"%lf\">100%% = %s</text>\n",
	    c_x0, y - c_fontindent, c_fontsize, SAB3(sumvalue));

    /* draw legend */
    if (legend) {	    
	c_barwidth = c_fontsize;
	c_ibspace = c_fontindent;
	xpos = x / 2.0 + c_barwidth;
	sum = 0.0;
	for (j = 0; j <= i; j++) {
	    if (j == i) {
		if (!(percent = 100.0 - sum)) break;
		label = llabel;
    		tmp = sumvalue * percent / 100.0;
	    } else {
		percent = icdata[j].value * 100.0 / sumvalue;
		sum += percent;
		label = icdata[j].label;
		tmp = icdata[j].value;
	    }
	    ypos = c_ibspace + (c_ibspace + c_barwidth) * (double)j;
	    SAfprintf(ostream, "\
\t\t<rect class=\"c%d\" x=\"%lf\" y=\"%lf\" width=\"%lf\" height=\"%lf\" stroke-width=\"%lf\" />\n\
\t\t<text class=\"stext\" x=\"%lf\" y=\"%lf\" font-size=\"%lf\">%s (%s) %.1lf%%</text>\n",
		    (j % 7) + 1, xpos, ypos, c_barwidth, c_barwidth, c_linewidth,
		    xpos + c_barwidth + c_ibspace, ypos + c_barwidth - c_fontindent, c_fontsize, label, SAB3(tmp), percent);
	    }
    }
    
    SAfcloseS(ostream);
}


void SSVGdrawbars(struct ChartData *icdata, const char relative);
void SSVGdrawbars(struct ChartData *icdata, const char relative) {
    FILE *ostream;
    int count, i = 0, j, k;
    double c_barwidth, c_linewidth, c_fontsize, c_yaxistep, c_ibspace, 
	   c_fontindent, c_halfbar, c_xshift, c_yshift, 
	   xpos, ysize, maxvalue, tmp;


    count = SSVGcountnmax(icdata, &maxvalue);
    /* compute operational values */
    y = x / YXRATIO;
    c_linewidth = y / LWDIV;						/* width of line		*/
    c_fontsize = x / FSDIV;						/* size of font			*/
    c_fontindent = c_fontsize * TINDENT;				/* indent of font from object	*/
    c_yshift = y - c_fontsize - c_fontindent * 2.0;			/* y axis shift			*/
    c_xshift = c_fontsize * FONTPROP + c_fontindent * 2.0;		/* x axis shift			*/
    c_ibspace = (BC_IBSPDIV + BC_IBSPDIV / (double)count);		/* inter-bar space		*/
    c_barwidth = (x - c_xshift) / ((double)count + c_ibspace * (double)(count + 1));	/* width of bar	*/
    c_halfbar = c_barwidth / 2.0;					/* half of bar width		*/

    ostream = SAfopenS();

    /* compute steps of lines at y axis */
    if (relative) for (i = 0; pow(10.0, (double)i) < maxvalue; i++);
    else i = 14;
    c_yaxistep = c_yshift / (double)(i + 1);
    /* draw y axis lines and labels */
    for (j = 0;; j++)
	for (k = 0; k < 3; k++) {
	    if ((tmp = pow(1024.0, (double)j) * pow(10.0, (double)k)) == 1.0) continue;
	    ysize = c_yshift - c_yaxistep * log10(tmp);
	    SAfprintf(ostream, "\
\t\t<path d=\"M%lf,%lf L%lf,%lf\" stroke-width=\"%lf\" />\n\
\t\t<text x=\"%lf\" y=\"%lf\" class=\"etext\" font-size=\"%lf\">%s</text>\n", 
		    c_xshift + c_linewidth, ysize, x, ysize, c_linewidth,
		    c_xshift - c_fontindent, ysize, c_fontsize, SAB3(tmp));
	    if (tmp > pow(10.0, (double)i)) goto end;
	}
end:
    /* draw bars with labels */
    for (i = 0; i < count; i++) {
	xpos = c_xshift + c_barwidth * (c_ibspace + (1.0 + c_ibspace) * (double)i);

	/* label */
	SAfprintf(ostream, "\
\t\t<text x=\"%lf\" y=\"%lf\" font-size=\"%lf\">%s</text>\n",
		  xpos + c_halfbar, c_yshift + c_fontindent + c_fontsize, c_fontsize, icdata[i].label);
	
	if (icdata[i].value) ysize = c_yaxistep * log10(icdata[i].value);
	else continue;
	
	/* bar and value on bar top */
	SAfprintf(ostream, "\
\t\t<rect x=\"%lf\" y=\"%lf\" width=\"%lf\" height=\"%lf\" stroke-width=\"%lf\" />\n\
\t\t<text x=\"%lf\" y=\"%lf\" font-size=\"%lf\">%s</text>\n",
	    xpos, c_yshift - ysize, c_barwidth, ysize, c_linewidth,
	    xpos + c_halfbar, c_yshift - ysize - c_fontindent, c_fontsize, SAB3(icdata[i].value));
    }
    
    /* draw x/y axes */
    SAfprintf(ostream, "\
\t\t<path d=\"M%lf,%lf L%lf,%lf L%lf,%lf\" class=\"aline\" stroke-width=\"%lf\" />\n", 
	    c_xshift, 0.0, c_xshift, c_yshift, x, c_yshift, c_linewidth * BC_AWCOEF);

    SAfcloseS(ostream);
}


void SVGReportAnyGraphs(double *mBytes, const char *vname, const char *name, short int elcount) {
    FILE *UGR;
    struct ChartData *UGRCD;
    char *UGRN;
    short int i, k = (elcount == 24) ? 0 : 1;


    UGR = SAfopenH("../", 'g', vname, NULL, 1, "%s/g_%s.html", icfg.ReportDir, name);
    
    UGRCD = (struct ChartData *)SAmalloc(((size_t)(elcount + 1)) * sizeof(struct ChartData));
    UGRN = (char *)SAmalloc(((size_t)elcount) * sizeof(char) * 3);
    for (i = 0; i < elcount; i++) {
	UGRCD[i].label = &UGRN[i * 3L];
	SAsnprintf(UGRCD[i].label, 3L, "%hd", i + k);
	UGRCD[i].value = mBytes[i];
    }
    UGRCD[i].label = NULL;

    SAsetlocale(LC_NUMERIC, "C");
    if (cfg.users_graphics_svg_pie) SSVGdrawpie(UGRCD, cfg.users_graphics_svg_mode);
    else SSVGdrawbars(UGRCD, cfg.users_graphics_svg_mode);
    SAsetlocale(LC_NUMERIC, cfg.locale);
    
    SAfprintf(UGR, "\
<tr><td class=el><object type=\"image/svg+xml\" data=\"%d.svg\" width=\"%.0f\" height=\"%.0f\"></object>",
	      cfg.users_graphics_svg - 1, x, y);
 
    SAfcloseH(UGR);
}
